/*********************************************************************
 *
 *
 **********************************************************************/
#include <stdio.h>
#include <asm/io.h>		/* port-io */
#include <sys/io.h> 		/* ioperm() (glibc) */
#include <time.h>		/* nanosleep() */

/*---------------------------------------------------------------------------
 * From src/linux/include/linux/lp.h:
 *
 * bit defines for 8255 status port
 * base + 1
 * accessed with LP_S(minor), which gets the byte...
 *
 * #define LP_PBUSY        0x80    inverted input, active high  
 * #define LP_PACK         0x40    unchanged input, active low  
 * #define LP_POUTPA       0x20    unchanged input, active high 
 * #define LP_PSELECD      0x10    unchanged input, active high 
 * #define LP_PERRORP      0x08    unchanged input, active low  
 *
 * defines for 8255 control port
 * base + 2 
 * accessed with LP_C(minor)
 *
 * #define LP_PINTEN       0x10   high to read data in or-ed with data out
 * #define LP_PSELECP      0x08   in verted output, active low          
 * #define LP_PINITP       0x04   unchanged output, active low          
 * #define LP_PAUTOLF      0x02   inverted output, active low           
 * #define LP_PSTROBE      0x01   short high output on raising edge      
 *---------------------------------------------------------------------------*/
#include <linux/lp.h>

static unsigned long 	PortBase = 0;

#define LP_DATA		(PortBase)
#define LP_STAT		(PortBase+1)
#define LP_CTRL		(PortBase+2)

#define D0		(1<<0)
#define D1		(1<<1)
#define D2		(1<<2)
#define D3		(1<<3)
#define D4		(1<<4)
#define D5		(1<<5)
#define D6		(1<<6)
#define D7		(1<<7)

#define	OUT_PORT	LP_DATA
#define IN_PORT		LP_STAT

#define EN_MOSI_LED_SCLK	D3
#define MOSI			D5
#define LED			D6
#define SCLK			D4

#define	EN_MISO_RES		D2
#define RESET			D7
#define	MISO			LP_PACK

#define TEST_OUT		D1
#define TEST_IN			LP_PBUSY		

/*-------------------- macros ----------------------------------------------*/
#define NANOSECS(a)		(a)
#define MICROSECS(a)		(1000ul*NANOSECS(a))
#define MILLISECS(a)		(1000ul*MICROSECS(a))

#define CMD_LEN(a)		(sizeof(a)/sizeof(a[0]))

/*-------------------- local headers ---------------------------------------*/
#include "at89s8252.h"
#include "debug.h"
#include "loaddata.h"

#ifdef extern
#  undef extern
#endif
#define extern
#  include "port-io.h"
#undef extern

/*-------------------- global variables ------------------------------------*/
static unsigned char OutState = 0;

#define MAX_PRG_EN_RETRY	32
#define PRG_CODE_ROM_RETRYS	32
#define PRG_DATA_ROM_RETRYS	32

/*-------------------- commands --------------------------------------------*/
typedef unsigned char BYTE_T;
typedef BYTE_T CMD_T;

static CMD_T CmdPrgEnable[] 	= { 0xAC, 0x53, 0x00 };
static CMD_T CmdChipErase[] 	= { 0xAC, 0x04, 0x00 };
static CMD_T CmdReadCodeMem[] 	= { 0x01, 0x00, 0x00 };
static CMD_T CmdWriteCodeMem[]	= { 0x02, 0x00, 0x00 };
static CMD_T CmdReadDataMem[] 	= { 0x05, 0x00, 0x00 };
static CMD_T CmdWriteDataMem[]	= { 0x06, 0x00, 0x00 };
static CMD_T CmdWriteLock[] 	= { 0xAC, 0x07, 0x00 };





/*-------------------- delay times -----------------------------------------*/
#define RES_WAIT		MILLISECS(1)
#define PRG_EN_WAIT		MILLISECS(1)
#define CHIP_ERASE_WAIT		MILLISECS(20)
/*#define BYTE_PRG_WAIT		NANOSECS(500)	*/
#define BYTE_PRG_WAIT		NANOSECS(0)

/*-------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int InitStk200(unsigned long Port) {
  int ErrCode;

  PortBase = Port;
  /*----------------- demanding port access --------------------------*/  
  DBG_PRINTF(DBG_IO, "ioperm 0x%03lX, 3 Bytes, On...", PortBase);
  ErrCode = ioperm(PortBase, 3, 1);
  DBG_PRINTF(DBG_IO, "-> %d\n", ErrCode);
  if(ErrCode != 0) { 
    PortBase = 0;
    return(-1);
  }
  /*----------------- initialize port --------------------------------*/
  outb(OutState = 0xFF, OUT_PORT);
  /*----------------- Check for stk200-adaptor -----------------------*/
  if((inb(IN_PORT) & TEST_IN) != 0) {		/* Busy is inverted! */
    DBG_PRINTF(DBG_IO, "stk200-adaptor not found (1)\n");
    return(-2);
  }  
  outb((OutState &= ~TEST_OUT), OUT_PORT);
  if((inb(IN_PORT) & TEST_IN) == 0) {		/* Busy is inverted! */
    DBG_PRINTF(DBG_IO, "stk200-adaptor not found (2)\n");
    return(-3);
  }  
  DBG_PRINTF(DBG_IO, "stk200-adaptor found!\n");
  return(0);
}

/*-------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ReleaseStk200(void) {
  int ErrCode;

  if(PortBase == 0) return(-1);
  /*----------------- reinitialize port ------------------------------*/
  outb(OutState = 0xFF, OUT_PORT);
  /*----------------- the end ----------------------------------------*/
  DBG_PRINTF(DBG_IO, "ioperm 0x%03lX, 3 Bytes, Off...", PortBase);
  ErrCode = ioperm(PortBase, 3, 0);
  DBG_PRINTF(DBG_IO, "-> %d\n", ErrCode);
  PortBase = 0;
  return(ErrCode);
}

/*-------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static void NanoDelay(unsigned long Del) {
  struct timespec Delay, Remaining; 
  int ErrCode;
  
  if(Del <= 0) return; 
  Delay.tv_sec	= 0;
  Delay.tv_nsec = Del;
  while(1){
    Remaining.tv_nsec = 0;
    ErrCode = nanosleep(&Delay, &Remaining);
    if(ErrCode == 0) break;    
    if(Remaining.tv_nsec == 0) {
      DBG_PRINTF(DBG_IO, "*** NanoDelay(%lu): Error calling nanosleep\n", Del);
      return;
    }
    Delay.tv_nsec = Remaining.tv_nsec;
  } /* while(1) */
}

/*-------------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static BYTE_T Stk200SendByte(CMD_T OutData) {
  BYTE_T InData;
  int i;
  
  InData = 0;
  DBG_PRINTF(DBG_IO, "Stk200SendByte(0x%02X):", (BYTE_T)OutData);
  for(i = 0; i < 8; i++) {
    InData <<= 1;
    if((OutData & (1<<7)) == 0) {
      outb((OutState &= ~MOSI), OUT_PORT);
      DBG_PRINTF(DBG_IO, "s0%c", i < 7 ? ',' : ' ');
    } else { 
      outb((OutState |= MOSI), OUT_PORT);
      DBG_PRINTF(DBG_IO, "s1%c", i < 7 ? ',' : ' ');
    }
    outb((OutState |= SCLK), OUT_PORT);
    if((inb(IN_PORT) & MISO) != 0){ 
      InData |= 1;
      DBG_PRINTF(DBG_IO, "r1%c", i < 7 ? ',' : ' '); 
    } else {
      DBG_PRINTF(DBG_IO, "r0%c", i < 7 ? ',' : ' ');
    }
    outb((OutState &= ~SCLK), OUT_PORT);
    OutData <<= 1;
  } /* for(i = 0; i < 8; i++) */
  DBG_PRINTF(DBG_IO, "->0x%02X\n", InData);
  return(InData);
} 

/*------------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static int SendPrgEnable(void) {
  int i, ErrCode;
  BYTE_T ReadBack;
  
  ErrCode = 0;
  DBG_PRINTF(DBG_IO, "SendPrgEnable():\n");
  for(i = 0; i < CMD_LEN(CmdPrgEnable); i++) {
    DBG_PRINTF(DBG_IO, "\tSendPrgEnable(): Sending 0x%02X\n\t", CmdPrgEnable[i]);    
    ReadBack = Stk200SendByte(CmdPrgEnable[i]);
#if 0
    if(i == 2) {
      if(ReadBack != CmdPrgEnable[i-1]) ErrCode = 1;
      DBG_PRINTF(DBG_IO, "\tSendPrgEnable(): Error, Sent 0x%02X, got 0x%02X\n",
      				CmdPrgEnable[i-1] & 0xFF, ReadBack);
    }
#endif
  } /* for(i = 0; i < CMD_LEN(PrgEnable); i++)  */
  return(ErrCode);
}

/*------------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static int PrgEnable(void) {
  int Retry;
  
  for(Retry = 0; Retry <= MAX_PRG_EN_RETRY; Retry++) {
    if(SendPrgEnable() == 0) { 
      DBG_PRINTF(DBG_IO, "PrgEnable() succesfull on %d. attempt\n", Retry+1);
      return(0);
    }
  }
  return(-1);
}

/*-------------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
int ActivateStk200(void) {
  outb((OutState &= ~RESET), OUT_PORT);
  outb((OutState &= ~EN_MISO_RES), OUT_PORT);
  outb((OutState &= ~(MOSI|LED|SCLK)), OUT_PORT);
  outb((OutState &= ~EN_MOSI_LED_SCLK), OUT_PORT);
  outb((OutState |= RESET), OUT_PORT);
  NanoDelay(RES_WAIT);
  outb((OutState &= ~RESET), OUT_PORT);
  outb((OutState |= RESET), OUT_PORT);
  NanoDelay(PRG_EN_WAIT);
  if(PrgEnable() != 0) {
    fprintf(stderr,"*** Error: cannot enable programming! ***\n");
    ReleaseStk200();
    exit(-3);
  }
  return(0);
}

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ChipErase(void) {
  Stk200SendByte(CmdChipErase[0]);
  Stk200SendByte(CmdChipErase[1]);
  Stk200SendByte(CmdChipErase[2]);
  NanoDelay(CHIP_ERASE_WAIT);  
  return(0);
} 


/*-------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static BYTE_T CodeHighAddr(long Address, DEVICE_T *Dev) { 
  unsigned long Mask;
  
  Mask = ((Dev->CodeSize) - 1) & ~0xFF;
  return((Address & Mask) >> 5);
}


/*-------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static BYTE_T DataHighAddr(long Address, DEVICE_T *Dev) { 
  unsigned long Mask;
  
  Mask = ((Dev->DataSize) - 1) & ~0xFF;
  return((Address & Mask) >> 5);
}

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static BYTE_T ReadCodeMemByte(long Address, DEVICE_T *Dev) {
  Stk200SendByte(CmdReadCodeMem[0] | CodeHighAddr(Address, Dev));
  Stk200SendByte((BYTE_T)(Address & 0xFF));    
  return(Stk200SendByte(CmdReadCodeMem[2]));
}

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ReadCodeMem(unsigned char *Buffer, long Len, DEVICE_T *Dev) {
  long Address;

  DBG_PRINTF(DBG_IO, "ReadCodeMem(%ld=%ldk)\n", Len, Len/1024);
  if(Len > Dev->CodeSize) {
    fprintf(stderr,"Warning ReadCodeMem(): "\
    		"demanded length %ld truncated to %ld\n", 
    		Len, Dev->CodeSize);
    Len = Dev->CodeSize;
  }
  for(Address = 0; Address < Len; Address++) {
    *Buffer++ = ReadCodeMemByte(Address, Dev);
  } /* for(Address = 0; Address < Len; Address++) */
  return(Len);
} 

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static int WriteCodeMemByte(long Address, BYTE_T Data, DEVICE_T *Dev) {
  BYTE_T VerifyData; 
  int Retry;

  Stk200SendByte(CmdWriteCodeMem[0] | CodeHighAddr(Address, Dev));
  Stk200SendByte((BYTE_T)(Address & 0xFF));
  Stk200SendByte(Data);
  NanoDelay(BYTE_PRG_WAIT); 
  for(Retry = 0; Retry < PRG_CODE_ROM_RETRYS; Retry++) {
    if((VerifyData = ReadCodeMemByte(Address, Dev)) == Data) { 
      DBG_PRINTF(DBG_IO, "WriteCodeMemByte(): needed %d retries for verifing\n",\
      		Retry + 1);
      return(0);
    } 
  }          
  fprintf(stderr,"WriteCodeMemByte(): Cannot verify Address %ld, "\
  		"wrote 0x%02X, read 0x%02X\n", Address, Data, VerifyData);
  return(-1); 
}

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int WriteCodeMem(unsigned char *Buffer, long Len, int DevErased, DEVICE_T *Dev) {
  long Address;

  DBG_PRINTF(DBG_IO, "WriteCodeMem(%ld=%ldk)\n", Len, Len/1024);
  if(Len > Dev->CodeSize) {
    fprintf(stderr,"Warning WriteCodeMem(): "\
    		"demanded length %ld truncated to %ld\n", 
    		Len, Dev->CodeSize);
    Len = Dev->CodeSize;
  }
  for(Address = 0; Address < Len; Address++) {
    if((DevErased == 0) || (*Buffer != BUFFER_INIT_VAL)) {	
      if(WriteCodeMemByte(Address, *Buffer, Dev) != 0) return(-1); 
    }  /* if((DevErased == 0) || (*Buffer != BUFFER_INIT_VAL) */
    Buffer++;
  } /* for(Address = 0; Address < Len; Address++) */
  return(Len);
}

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static BYTE_T ReadDataMemByte(long Address, DEVICE_T *Dev) {
  Stk200SendByte(CmdReadDataMem[0] | DataHighAddr(Address, Dev));
  Stk200SendByte((BYTE_T)(Address & 0xFF));    
  return(Stk200SendByte(CmdReadDataMem[2]));
}

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ReadDataMem(unsigned char *Buffer, long Len, DEVICE_T *Dev) {
  long Address;

  DBG_PRINTF(DBG_IO, "ReadDataMem(%ld=%ldk)\n", Len, Len/1024);
  if(Len > Dev->DataSize) {
    fprintf(stderr,"Warning ReadDataMem(): "\
    		"demanded length %ld truncated to %ld\n", 
    		Len, Dev->DataSize);
    Len = Dev->DataSize;
  }
  for(Address = 0; Address < Len; Address++) {
    *Buffer++ = ReadDataMemByte(Address, Dev);
  } /* for(Address = 0; Address < Len; Address++) */
  return(Len);
} 

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static int WriteDataMemByte(long Address, BYTE_T Data, DEVICE_T *Dev) {
  BYTE_T VerifyData; 
  int Retry;
  
  Stk200SendByte(CmdWriteDataMem[0] | DataHighAddr(Address, Dev));
  Stk200SendByte((BYTE_T)(Address & 0xFF));
  Stk200SendByte(Data);
  NanoDelay(BYTE_PRG_WAIT); 
  for(Retry = 0; Retry < PRG_DATA_ROM_RETRYS; Retry++) {
    if((VerifyData = ReadDataMemByte(Address, Dev)) == Data) { 
      DBG_PRINTF(DBG_IO, "WriteDataMemByte(): needed %d retries for verifing\n",\
      		Retry + 1);
      return(0);
    } 
  }          
  fprintf(stderr,"WriteDataMemByte(): Cannot verify Address %ld, "\
  		"wrote 0x%02X, read 0x%02X\n", Address, Data, VerifyData);
  return(-1); 
}

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int WriteDataMem(unsigned char *Buffer, long Len, int DevErased, DEVICE_T *Dev) {
  long Address;

  DBG_PRINTF(DBG_IO, "WriteDataMem(%ld=%ldk)\n", Len, Len/1024);
  if(Len > Dev->DataSize) {
    fprintf(stderr,"Warning WriteDataMem(): "\
    		"demanded length %ld truncated to %ld\n", 
    		Len, Dev->DataSize);
    Len = Dev->DataSize;
  }
  for(Address = 0; Address < Len; Address++) {
    if((DevErased == 0) || (*Buffer != BUFFER_INIT_VAL)) {
      if(WriteDataMemByte(Address, *Buffer, Dev) != 0) return(-1); 
    } /* if((DevErased == 0) || (*Buffer != BUFFER_INIT_VAL)) */
    Buffer++;
  } /* for(Address = 0; Address < Len; Address++) */
  return(Len);
}

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int WriteLockBits(LOCKBIT_T **Bits, DEVICE_T *Dev) {
  BYTE_T LockBitMask;
  int i;
  
  DBG_PRINTF(DBG_IO, "WriteLockBits(): ");
  LockBitMask = 0;
  for(i = 0; Dev->LockBits[i]!= NULL; i++) LockBitMask |= LockBits[i]->Mask;
  for(i = 0; Bits[i] != NULL; i++) LockBitMask &= ~(Bits[i]->Mask);
  DBG_PRINTF(DBG_IO, "Mask=0x%02X\n", LockBitMask);
  Stk200SendByte(CmdWriteLock[0]);
  Stk200SendByte(CmdWriteLock[1]|LockBitMask);
  Stk200SendByte(CmdWriteLock[2]);
  return(0);
}
