/*****************************************************************
 * mprog-eprommer
 *
 * File: loaddata.c
 * 
 *******************************************************************/
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#ifdef extern
#  undef extern
#endif
#define extern
#  include "loaddata.h"
#undef extern

#include "debug.h"
/*-------------------------- Constants ----------------------------*/
#define BUF_LEN			512
#define MAX_SREC_LEN		0xFF

/*-------------------------- Variables ----------------------------*/
static char LineBuffer[BUF_LEN];

#define IHEX_START	':'


/*---------------------------------------------------------------
 *
 *-----------------------------------------------------------------*/
static long LoadBin(FILE *Fp, unsigned long Offset, unsigned long Size, 
	unsigned char *Buffer, int Skip)			{
  int c;
  int ErrCode;
  unsigned long InpCount = 0, BuffCount = 0;
    
  DBG_PRINTF(DBG_LOAD, "LoadBin(Ofs=%lu, Size=%lu, Skip=%d", Offset, Size, Skip);
  if(Offset > 0) {
    if((ErrCode = fseek(Fp, (long)Offset, SEEK_SET))!= 0) {
      fprintf(stderr,"*** Error LoadBin(): invalid Offset (%lu) ***\n", Offset);
      return(-2);
    }
  }
  while(1){
    InpCount++;
    if((c = fgetc(Fp)) == EOF) { 
      DBG_PRINTF(DBG_LOAD, "\tGot EOF: Loaded %lu Bytes\n", BuffCount);
      return(BuffCount);
    }
    if(Skip > 1) if(InpCount % Skip != 0) continue;
    *Buffer++ = (unsigned char)c;
    BuffCount++;
    if(BuffCount > Size) { 
      DBG_PRINTF(DBG_LOAD, "\tReached Size: Loaded %lu Bytes\n", BuffCount);
      return(BuffCount);
    }
  }
  DBG_PRINTF(DBG_LOAD, "\tLoaded %lu Bytes\n", BuffCount);
  return(BuffCount);
}

/*---------------------------------------------------------------
 *
 *-------------------------------------------------------------------*/
static int HexNibble(char *Ch) {
  int ChU;
  if((Ch == NULL) || (*Ch == '\0')) return(-1);
  if(*Ch < '0') return(-1);
  if(*Ch <= '9') return(*Ch - '0');
  ChU = toupper(*Ch);
  if(ChU < 'A') return(-1);
  if(ChU <= 'F') return(ChU - 'A' + 10);
  return(-1);
}

/*---------------------------------------------------------------
 *
 *-------------------------------------------------------------------*/
static char *HexConv(int NumBytes, unsigned long *Data, 	\
				unsigned char *ChkSum, char *Line) {
  int TmpData, i;
  unsigned char Byte;
  
  if(NumBytes > 4) return(NULL);
  if(Data == NULL) return(NULL); 
  *Data = 0;
  for(i = 0; i < NumBytes; i++) {
    *Data <<= 8;
    if((TmpData = HexNibble(Line++)) < 0) return(NULL);
    Byte = TmpData << 4; 
    if((TmpData = HexNibble(Line++)) < 0) return(NULL);
    Byte |= TmpData;
    if(ChkSum) *ChkSum += Byte;
    *Data |= Byte;
  }
  return(Line);
}

/*--------------------------------------------------------------- 
 *
 *-----------------------------------------------------------------*/
static int SRecord(int RecType, char *Line, int *Len, \
			unsigned long *Addr, unsigned char *Data) {
  unsigned char ChkSum = 0;
  int i, NumBytes;
  unsigned long TmpData;
  int AddrLen[] = {0, 2, 3, 4 }; 

  if((RecType <= 0) || (RecType >= sizeof(AddrLen)/sizeof(AddrLen[0]))) {
    DBG_PRINTF(DBG_LOAD, "Invalid record type %d\n", RecType);
    return(-10);
  }
  DBG_PRINTF(DBG_LOAD, "SRecord(AddrLen=%d): ", AddrLen[RecType]);  
  if((Line = HexConv(1, &TmpData, &ChkSum, Line)) == NULL) { 
    DBG_PRINTF(DBG_LOAD, "Invalid record length\n");
    return(-10);
  }
  *Len = NumBytes = TmpData - 1 - AddrLen[RecType];
  if((Line = HexConv(AddrLen[RecType], Addr, &ChkSum, Line)) == NULL) { 
    DBG_PRINTF(DBG_LOAD, "Invalid record address\n");
    return(-11);
  }
  DBG_PRINTF(DBG_LOAD, "Len=%d,Addr=0x%08lX\n  Data(hex)  :",*Len, *Addr);
  for(i = 0; i < NumBytes; i++) {
    if((Line = HexConv(1, &TmpData, &ChkSum, Line)) == NULL) { 
      DBG_PRINTF(DBG_LOAD, "Invalid record data\n");
      return(-12);
    }
    *Data = (unsigned char)(TmpData & 0xFF);
    DBG_PRINTF(DBG_LOAD, "%02X",*Data);
    DBG_COND_PRINTF(DBG_LOAD, i < NumBytes - 1, ",");
    Data++;
  }
  DBG_PRINTF(DBG_LOAD, "\n  ChkSum(hex):%02X",ChkSum);
  if((Line = HexConv(1, &TmpData, &ChkSum, Line)) == NULL) { 
    DBG_PRINTF(DBG_LOAD, "Invalid check sum\n");
    return(-13);
  }
  DBG_PRINTF(DBG_LOAD, "->%02X\n",ChkSum);
  if(ChkSum != 0xFF) {
    fprintf(stderr,"*** Check-Sum-Error ***\n");
    return(-14);
  }
  return(RecType);
}

/*--------------------------------------------------------------- 
 *
 *-----------------------------------------------------------------*/
static int SRecConv(char *Line, int *Len, unsigned long *Addr, unsigned char *Data) {
  
  if((Line == NULL) || (*Line == '\0')) return(-1);
  if(toupper(*Line) != 'S') return(-2);
  DBG_PRINTF(DBG_LOAD, "S-Rec:%s", Line);
  Line++;
  *Len  = 0;
  *Addr = 0;
  switch(*Line) {
    case '0':	return(0);
    case '1': 	return(SRecord(1, Line+1, Len, Addr, Data));
    case '2': 	return(SRecord(2, Line+1, Len, Addr, Data));
    case '3': 	return(SRecord(3, Line+1, Len, Addr, Data));
    case '5': 	return(5);
    case '7':	return(7);
    case '8':	return(8);
    case '9': 	return(9);
    default:	return(-3);
  }
  return(-3);
}

/*---------------------------------------------------------------
 *
 *-----------------------------------------------------------------*/
static int LoadSRec(FILE *Fp, unsigned long Offset, unsigned long Size, 
	unsigned char *Buffer, int Skip)			{
  char *Line;
  unsigned char *BuffPtr, *DataPtr;
  unsigned char Data[MAX_SREC_LEN];
  unsigned long BuffCount = 0;
  unsigned long Address, BufAddr;
  unsigned long MinAddr, MaxAddr;
  int RecLen, DataOffset, DataLen, RecType, i;
  int LastDataIdx = 0;
  long LineCount = 0;
  
  if(Skip < 1) { 
    fprintf(stderr,"*** Error LoadSRec(): Skip invalid (%d)\n", Skip);
    return(-1);
  }
  MinAddr = 0xFFFFFFFFul;
  MaxAddr = 0;
  while(1) {
    LineCount++;
    if((Line = fgets(LineBuffer, BUF_LEN, Fp)) == NULL) return(MaxAddr - MinAddr);
    if((RecType = SRecConv(Line, &RecLen, &Address, Data)) < 0) {
      fprintf(stderr,"*** Error: Invalid S-Record in line %ld (Code %d)***\n", 
      		LineCount, RecType);
      return(-1);
    }
    DBG_PRINTF(DBG_LOAD, "Line %ld: S%d-Record\n", LineCount, RecType);
    if((RecType != 1) && (RecType != 2) && (RecType != 3)) continue;
    BufAddr = 0;
    DataOffset = DataLen = 0;
    if((Address + RecLen < Offset) || (Address > Offset + Size * Skip)) { 
      DBG_PRINTF(DBG_LOAD, "Record outside Buffer\n");
      continue;
    } 
    if(Address <= Offset) {
      BufAddr 	= 0;
      DataOffset= Offset - Address;
      DataLen	= RecLen - DataOffset; 
    } else {
      BufAddr = (Address - Offset) / Skip;
      DataLen = (BufAddr + RecLen > Size) ? Size - BufAddr : RecLen; 
    } 
    if(BufAddr < MinAddr) MinAddr = BufAddr;
    if((BufAddr + (DataLen / Skip)) > MaxAddr) MaxAddr = BufAddr + (DataLen / Skip); 
    DBG_PRINTF(DBG_LOAD, "  BuffAddr=0x%lX,Doffs=%d,Dlen=%d,D=",BufAddr, DataOffset, DataLen);
    BuffPtr = Buffer + BufAddr;
    DataPtr = Data + DataOffset;
    BuffCount += DataLen;
    for(i = 0; i < DataLen; i++) {
      if((*BuffPtr != BUFFER_INIT_VAL) && (*BuffPtr != *DataPtr)){
        fprintf(stderr,"warning line %ld: overwriting data (Addr.:0x%lX, old:0x%02X,new:0x%02X)\n",
		LineCount, Address+i+DataOffset, *BuffPtr, *DataPtr); 
      }
      if((Skip == 1) || ((i + LastDataIdx) % Skip == 0)) { 
        DBG_PRINTF(DBG_LOAD, "%02X%c", *DataPtr, i < DataLen - 1 ? ',' : '\n');
      	*BuffPtr++ = *DataPtr;
      }
      DataPtr++;
    } /* for(i = 0; i < DataLen; i++) */
    DBG_PRINTF(DBG_LOAD, "\n");
    if(Skip != 1) LastDataIdx = DataLen % Skip;
  } /* while(1) */
  return(MaxAddr - MinAddr);
}

/*--------------------------------------------------------------
 *
 *-----------------------------------------------------------------*/
static int IHexConv(char *Line, long *Address, long *Type, unsigned char *Data) {
  long NumberBytes, i;
  unsigned char ChkSum;
  long TempData;
  
  ChkSum = 0;
  if(*Line++ != IHEX_START) return(-1);
  if((Line = HexConv(1, &NumberBytes, &ChkSum, Line)) == NULL) {
    DBG_PRINTF(DBG_LOAD, "Invalid data number\n");
    return(-2);
  }
  if((Line = HexConv(2, Address, &ChkSum, Line)) == NULL) {
    DBG_PRINTF(DBG_LOAD, "Invalid Address\n");
    return(-3);
  }
  if((Line = HexConv(1, Type, &ChkSum, Line)) == NULL) {
    DBG_PRINTF(DBG_LOAD, "Invalid Type\n");
    return(-4);
  }
  for(i = 0; i < NumberBytes; i++) {
    if((Line = HexConv(1, &TempData, &ChkSum, Line)) == NULL) {
      DBG_PRINTF(DBG_LOAD, "Invalid Data\n");
      return(-5);
    }
    *Data++ = (unsigned char)(TempData & 0xFF);
  } /* for(i = 0; i < NumberBytes; i++) */
  if((Line = HexConv(1, &TempData, &ChkSum, Line)) == NULL) {
    DBG_PRINTF(DBG_LOAD, "Checksum error\n");
    return(-6);
  }
  if(ChkSum != 0) {
    DBG_PRINTF(DBG_LOAD, "Invalid Checksum\n");
    return(-7);
  }
  return((int)NumberBytes);
} 
  
/*---------------------------------------------------------------
 *
 *-----------------------------------------------------------------*/
static int LoadIHex(FILE *Fp, unsigned long Offset, unsigned long Size, 
	unsigned char *Buffer, int Skip)			{
  char *Line;
  unsigned char *BuffPtr, *DataPtr;
  unsigned char Data[MAX_SREC_LEN];
  unsigned long BuffCount = 0;
  unsigned long Address, BufAddr;
  unsigned long MinAddr, MaxAddr;
  int RecLen, DataOffset, DataLen, i;
  int LastDataIdx = 0;
  long LineCount = 0;
  long RecType;
  
  MinAddr = 0xFFFFFFFFul;
  MaxAddr = 0;
  if(Skip < 1) { 
    fprintf(stderr,"*** Error LoadIHex(): Skip invalid (%d)\n", Skip);
    return(-1);
  }
  while(1) {
    LineCount++;
    if((Line = fgets(LineBuffer, BUF_LEN, Fp)) == NULL) return(MaxAddr - MinAddr);
    if((RecLen = IHexConv(Line, &Address, &RecType, Data))  < 0) {
      fprintf(stderr,"*** Error: Invalid Record in line %ld (Code %d)***\n", 
      		LineCount, RecLen);
      return(-1);
    }
    DBG_PRINTF(DBG_LOAD, "Line %ld: Type=%ld\n", LineCount, RecType);
    BufAddr = 0;
    DataOffset = DataLen = 0;
    if((Address + RecLen < Offset) || (Address > Offset + Size * Skip)) { 
      DBG_PRINTF(DBG_LOAD, "Record outside Buffer\n");
      continue;
    } 
    if(Address <= Offset) {
      BufAddr 	= 0;
      DataOffset= Offset - Address;
      DataLen	= RecLen - DataOffset; 
    } else {
      BufAddr = (Address - Offset) / Skip;
      DataLen = (BufAddr + RecLen > Size) ? Size - BufAddr : RecLen; 
    } 
    if(BufAddr < MinAddr) MinAddr = BufAddr;
    if((BufAddr + (DataLen / Skip)) > MaxAddr) MaxAddr = BufAddr + (DataLen / Skip);
    DBG_PRINTF(DBG_LOAD, "  BuffAddr=0x%lX,Doffs=%d,Dlen=%d,D=",BufAddr, DataOffset, DataLen);
    BuffPtr = Buffer + BufAddr;
    DataPtr = Data + DataOffset;
    BuffCount += DataLen;
    for(i = 0; i < DataLen; i++) {
      if((*BuffPtr != BUFFER_INIT_VAL) && (*BuffPtr != *DataPtr)){
        fprintf(stderr,"warning line %ld: overwriting data (Addr.:0x%lX, old:0x%02X,new:0x%02X)\n",
		LineCount, Address+i+DataOffset, *BuffPtr, *DataPtr); 
      }
      if((Skip == 1) || ((i + LastDataIdx) % Skip == 0)) { 
        DBG_PRINTF(DBG_LOAD, "%02X%c", *DataPtr, i < DataLen - 1 ? ',' : '\n');
      	*BuffPtr++ = *DataPtr;
      }
      DataPtr++;
    } /* for(i = 0; i < DataLen; i++) */
    DBG_PRINTF(DBG_LOAD, "\n");
    if(Skip != 1) LastDataIdx = DataLen % Skip;
  } /* while(1) */
  return(MaxAddr - MinAddr);
}

/*---------------------------------------------------------------
 *
 *-----------------------------------------------------------------*/
long LoadData(FILE *Fp, DATA_T Type,			\
		unsigned long Offset, unsigned long Size, 	\
		unsigned char *Buffer, int Skip) 		{
  int ErrCode;
  
  if(Fp == NULL) return(-1);
  switch(Type){
    case BINARY: {
    	ErrCode = LoadBin(Fp, Offset, Size, Buffer, Skip);
      } break;
    case MOTO_SREC: {
    	ErrCode = LoadSRec(Fp, Offset, Size, Buffer, Skip);
      } break;
    case INTEL_HEX: {
    	ErrCode = LoadIHex(Fp, Offset, Size, Buffer, Skip);
      } break;
    default : {
        fprintf(stderr,"*** Invalid Datatype (%d) ***\n", Type);
        ErrCode = -1;
      } break;
  }
  fclose(Fp);
  return(ErrCode); 
}
