/*********************************************************************
 *
 *
 **********************************************************************/

#include <stdio.h>
#include <asm/io.h>		/* port-io */
#include <sys/io.h> 		/* ioperm() (glibc) */
#include <time.h>		/* nanosleep() */

/*---------------------------------------------------------------------------
 * From src/linux/include/linux/lp.h:
 *
 * bit defines for 8255 status port
 * base + 1
 * accessed with LP_S(minor), which gets the byte...
 *
 * #define LP_PBUSY        0x80    inverted input, active high  
 * #define LP_PACK         0x40    unchanged input, active low  
 * #define LP_POUTPA       0x20    unchanged input, active high 
 * #define LP_PSELECD      0x10    unchanged input, active high 
 * #define LP_PERRORP      0x08    unchanged input, active low  
 *
 * defines for 8255 control port
 * base + 2 
 * accessed with LP_C(minor)
 *
 * #define LP_PINTEN       0x10   high to read data in or-ed with data out
 * #define LP_PSELECP      0x08   in verted output, active low          
 * #define LP_PINITP       0x04   unchanged output, active low          
 * #define LP_PAUTOLF      0x02   inverted output, active low           
 * #define LP_PSTROBE      0x01   short high output on raising edge      
 *---------------------------------------------------------------------------*/
#include <linux/lp.h>

static unsigned long		PortBase = 0;

#define LP_DATA			(PortBase)
#define LP_STAT			(PortBase+1)
#define LP_CTRL			(PortBase+2)

#define D0			(1<<0)
#define D1			(1<<1)
#define D2			(1<<2)
#define D3			(1<<3)
#define D4			(1<<4)
#define D5			(1<<5)
#define D6			(1<<6)
#define D7			(1<<7)

#define	OUT_PORT		LP_DATA
#define IN_PORT			LP_STAT

#define EN_MOSI_LED_SCLK	D3
#define MOSI			D5
#define LED			D6
#define SCLK			D4

#define	EN_MISO_RES		D2
#define RESET			D7
#define	MISO			LP_PACK

#define TEST_OUT		D1
#define TEST_IN			LP_PBUSY		

#include "avr.h"
/*-------------------- types -----------------------------------------------*/
typedef unsigned char BYTE_T;
typedef BYTE_T CMD_T;

typedef struct sLockBits {	LOCKBIT_T	Lb;
				BYTE_T		Mask;
		} LOCK_T;	

typedef struct sFuseBits {	FUSEBIT_T	Fs;
				BYTE_T		Mask;
		} FUSE_T;	


/*-------------------- macros ----------------------------------------------*/
#define NANOSECS(a)		a
#define MICROSECS(a)		(1000ul*NANOSECS(a))
#define MILLISECS(a)		(1000ul*MICROSECS(a))

#define CMD_LEN(a)		(sizeof(a)/sizeof(a[0]))

/*-------------------- constants -------------------------------------------*/
#define MAX_PRG_EN_RETRY	32
#define PRG_CODE_ROM_RETRYS	32
#define PRG_DATA_ROM_RETRYS	32


#define SIG_LEN			3

#define CODE_RD_HIGH_MASK	(1<<3)
#define CODE_WR_HIGH_MASK	(1<<3)

/*-------------------- delay times -----------------------------------------*/
#define RES_WAIT	MICROSECS(10)
#define PRG_EN_WAIT	MILLISECS(20)
#define CHIP_ERASE_WAIT	MILLISECS(18)
#define BYTE_PRG_WAIT	NANOSECS(0)


/*-------------------- local headers ---------------------------------------*/
#include "debug.h"

#ifdef extern
#  undef extern
#endif
#define extern
#  include "port-io.h"
#undef extern

#include "loaddata.h"
/*-------------------- global variables ------------------------------------*/
static unsigned char OutState = 0;

/*-------------------- commands --------------------------------------------*/

static CMD_T CmdPrgEnable[] 	= { 0xAC, 0x53, 0x00, 0x00 };
static CMD_T CmdChipErase[] 	= { 0xAC, 0x80, 0x00, 0x00 };
static CMD_T CmdReadCodeMem[] 	= { 0x20, 0x00, 0x00, 0x00 };
static CMD_T CmdWriteCodeMem[] 	= { 0x40, 0x00, 0x00, 0x00 };
static CMD_T CmdReadDataMem[] 	= { 0xA0, 0x00, 0x00, 0x00 };
static CMD_T CmdWriteDataMem[] 	= { 0xC0, 0x00, 0x00, 0x00 };
static CMD_T CmdReadLock[]	= { 0x58, 0x00, 0x00, 0x00 };
static CMD_T CmdWriteLock[] 	= { 0xAC, 0xFF, 0x00, 0x00 };
static CMD_T CmdWriteFuse[] 	= { 0xAC, 0xBF, 0x00, 0x00 };
static CMD_T CmdReadFuse[]	= { 0x50, 0x00, 0x00, 0x00 };
static CMD_T CmdReadSig[] 	= { 0x30, 0x00, 0x00, 0x00 };

/*-------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int InitStk200(unsigned long Port) {
  int ErrCode;

  PortBase = Port;
  /*----------------- demanding port access --------------------------*/  
  DBG_PRINTF(DBG_IO, "ioperm 0x%03lX, 3 Bytes, On...", PortBase);
  ErrCode = ioperm(PortBase, 3, 1);
  DBG_PRINTF(DBG_IO, "-> %d\n", ErrCode);
  if(ErrCode != 0) return(-1);
  /*----------------- initialize port --------------------------------*/
  outb(OutState = 0xFF, OUT_PORT);
  /*----------------- Check for stk200-adaptor -----------------------*/
  if((inb(IN_PORT) & TEST_IN) != 0) {		/* Busy is inverted! */
    DBG_PRINTF(DBG_IO, "stk200-adaptor not found (1)\n");
    return(-2);
  }  
  outb((OutState &= ~TEST_OUT), OUT_PORT);
  if((inb(IN_PORT) & TEST_IN) == 0) {		/* Busy is inverted! */
    DBG_PRINTF(DBG_IO, "stk200-adaptor not found (2)\n");
    return(-3);
  }  
  DBG_PRINTF(DBG_IO, "stk200-adaptor found!\n");
  return(0);
}

/*-------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ReleaseStk200(void) {
  int ErrCode;

  if(PortBase == 0) return(-1);
  /*----------------- reinitialize port ------------------------------*/
  outb(OutState = 0xFF, OUT_PORT);
  /*----------------- the end ----------------------------------------*/
  DBG_PRINTF(DBG_IO, "ioperm 0x%03lX, 3 Bytes, Off...", PortBase);
  ErrCode = ioperm(PortBase, 3, 0);
  DBG_PRINTF(DBG_IO, "-> %d\n", ErrCode);
  PortBase = 0;
  return(ErrCode);
}

/*-------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static void NanoDelay(unsigned long Del) {
  struct timespec Delay, Remaining; 
  int ErrCode;
  
  if(Del <= 0) return; 
  Delay.tv_sec	= 0;
  Delay.tv_nsec = Del;
  while(1){
    Remaining.tv_nsec = 0;
    ErrCode = nanosleep(&Delay, &Remaining);
    if(ErrCode == 0) break;    
    if(Remaining.tv_nsec == 0) {
      DBG_PRINTF(DBG_IO, "*** NanoDelay(%lu): Error calling nanosleep\n", Del);
      return;
    }
    Delay.tv_nsec = Remaining.tv_nsec;
  } /* while(1) */
}

/*-------------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static BYTE_T Stk200SendByte(CMD_T OutData) {
  BYTE_T InData;
  int i;
  
  InData = 0;
  DBG_PRINTF(DBG_IO, "Stk200SendByte(0x%02X):", (BYTE_T)OutData);
  for(i = 0; i < 8; i++) {
    InData <<= 1;
    if((OutData & (1<<7)) == 0) {
      outb((OutState &= ~MOSI), OUT_PORT);
      DBG_PRINTF(DBG_IO, "s0%c", i < 7 ? ',' : ' ');
    } else { 
      outb((OutState |= MOSI), OUT_PORT);
      DBG_PRINTF(DBG_IO, "s1%c", i < 7 ? ',' : ' ');
    }
    outb((OutState |= SCLK), OUT_PORT);
    if((inb(IN_PORT) & MISO) != 0){ 
      InData |= 1;
      DBG_PRINTF(DBG_IO, "r1%c", i < 7 ? ',' : ' '); 
    } else {
      DBG_PRINTF(DBG_IO, "r0%c", i < 7 ? ',' : ' ');
    }
    outb((OutState &= ~SCLK), OUT_PORT);
    OutData <<= 1;
    NanoDelay(0);
  } /* for(i = 0; i < 8; i++) */
  DBG_PRINTF(DBG_IO, "->0x%02X\n", InData);
  return(InData);
} 

/*------------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static int SendPrgEnable(void) {
  int i, ErrCode;
  BYTE_T ReadBack;
  
  ErrCode = 0;
  DBG_PRINTF(DBG_IO, "SendPrgEnable():\n");
  for(i = 0; i < CMD_LEN(CmdPrgEnable); i++) {
    DBG_PRINTF(DBG_IO, "\tSendPrgEnable(): Sending 0x%02X\n\t", CmdPrgEnable[i]);    
    ReadBack = Stk200SendByte(CmdPrgEnable[i]);
    if(i == 2) {
      if(ReadBack != CmdPrgEnable[i-1]) ErrCode = 1;
      DBG_PRINTF(DBG_IO, "\tSendPrgEnable(): Error, Sent 0x%02X, got 0x%02X\n",
      				CmdPrgEnable[i-1] & 0xFF, ReadBack);
    }
  } /* for(i = 0; i < CMD_LEN(PrgEnable); i++)  */
  return(ErrCode);
}

/*------------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static int PrgEnable(void) {
  int Retry;
  
  for(Retry = 0; Retry <= MAX_PRG_EN_RETRY; Retry++) {
    if(SendPrgEnable() == 0) { 
      DBG_PRINTF(DBG_IO, "PrgEnable() succesfull on %d. attempt\n", Retry+1);
      return(0);
    }
  }
  return(-1);
}

/*-------------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
int ActivateStk200(void) {
  outb((OutState &= ~RESET), OUT_PORT);
  outb((OutState &= ~EN_MISO_RES), OUT_PORT);
  outb((OutState &= ~(MOSI|LED|SCLK)), OUT_PORT);
  outb((OutState &= ~EN_MOSI_LED_SCLK), OUT_PORT);
  outb((OutState |= RESET), OUT_PORT);
  NanoDelay(RES_WAIT);
  outb((OutState &= ~RESET), OUT_PORT);
  NanoDelay(PRG_EN_WAIT);
  if(PrgEnable() != 0) {
    fprintf(stderr,"*** Error: cannot enable programming! ***\n");
    ReleaseStk200();
    exit(-3);
  }
  return(0);
}

/*-------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static BYTE_T ReadSigByte(BYTE_T Addr) {

  DBG_PRINTF(DBG_IO, "ReadSigByte(%d)\n\t", Addr);
  Stk200SendByte(CmdReadSig[0]);
  Stk200SendByte(CmdReadSig[1]);
  Stk200SendByte(CmdReadSig[2] | Addr);
  return(Stk200SendByte(CmdReadSig[3]));
}

/*-------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ReadSignature(unsigned char *Sig, int Len) {
  int i;
  
  DBG_PRINTF(DBG_IO, "ReadSignature()\n\t->");
  for(i = 0; i < SIG_LEN; i++) { 
    Sig[i] = ReadSigByte((BYTE_T)i);
    DBG_PRINTF(DBG_IO, "0x%02X%c", Sig[i], (i < SIG_LEN-1)? ',' : '\n');
    if(i+1 == Len) return(Len);
  }
  return(SIG_LEN);
}

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ChipErase(void) {
  Stk200SendByte(CmdChipErase[0]);
  Stk200SendByte(CmdChipErase[1]);
  Stk200SendByte(CmdChipErase[2]);
  Stk200SendByte(CmdChipErase[3]);
  NanoDelay(CHIP_ERASE_WAIT);
  outb((OutState |= RESET), OUT_PORT);
  NanoDelay(RES_WAIT);
  return(ActivateStk200());
}

/*-------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static BYTE_T ReadCodeMemByte(long Address) {

  if((Address & 0x01) == 0) {
    Stk200SendByte(CmdReadCodeMem[0]);
  } else {
    Stk200SendByte(CmdReadCodeMem[0]|CODE_RD_HIGH_MASK);
  }
  Stk200SendByte(CmdReadCodeMem[1]|((Address >> 9) & 0xFF));
  Stk200SendByte(CmdReadCodeMem[2]|((Address >> 1) & 0xFF));
  return(Stk200SendByte(CmdReadCodeMem[3]));
}

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ReadCodeMem(unsigned char *Buffer, long Len) {
  long Address;
  DBG_PRINTF(DBG_IO, "ReadCodeMem(%ld=%ldk)\n", Len, Len/1024);
  for(Address = 0; Address < Len; Address++) { 
    DBG_PRINTF(DBG_IO, "\nAddress 0x%04lX\n", Address);
    *Buffer++ = ReadCodeMemByte(Address);
  }
  return(Len);
}

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static int WriteCodeMemByte(long Address, BYTE_T Data) {
  BYTE_T VerifyData; 
  int Retry;
  
  if((Address & 0x01) == 0) {
    Stk200SendByte(CmdWriteCodeMem[0]);
  } else {
    Stk200SendByte(CmdWriteCodeMem[0]|CODE_WR_HIGH_MASK);
  }
  Stk200SendByte(CmdWriteCodeMem[1]|((Address >> 9) & 0xFF));
  Stk200SendByte(CmdWriteCodeMem[2]|((Address >> 1) & 0xFF));
  Stk200SendByte(Data);
  NanoDelay(BYTE_PRG_WAIT); 
  for(Retry = 0; Retry < PRG_CODE_ROM_RETRYS; Retry++) {
    if((VerifyData = ReadCodeMemByte(Address)) == Data) { 
      DBG_PRINTF(DBG_IO, "WriteCodeMemByte(): needed %d retries for verifing\n",\
      		Retry + 1);
      return(0);
    }
  }
  fprintf(stderr,"WriteCodeMemByte(): Cannot verify Address %ld, "\
  		"wrote 0x%02X, read 0x%02X\n", Address, Data, VerifyData);
  return(-1); 
}
      
/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int WriteCodeMem(unsigned char *Buffer, long Len) {
  long Address;
  DBG_PRINTF(DBG_IO, "WriteCodeMem(%ld=%ldk)\n", Len, Len/1024);
  for(Address = 0; Address < Len; Address++) { 
    DBG_PRINTF(DBG_IO, "\nAddress 0x%04lX\n", Address);
    if(*Buffer != BUFFER_INIT_VAL) {
      if(WriteCodeMemByte(Address, *Buffer) != 0) return(-1); 
    } /* if(*Buffer != BUFFER_INIT_VAL) */
    Buffer++;
  } /* for(Address = 0; Address < Len; Address++) */
  return(Len);
}

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static BYTE_T ReadDataMemByte(long Address) {
  Stk200SendByte(CmdReadDataMem[0]);
  Stk200SendByte(CmdReadDataMem[1]|((Address >> 8) & 0xFF));
  Stk200SendByte(CmdReadDataMem[2]|( Address       & 0xFF));
  return(Stk200SendByte(CmdReadDataMem[3]));
}

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ReadDataMem(unsigned char *Buffer, long Len) {
  long Address;
  DBG_PRINTF(DBG_IO, "ReadDataMem(%ld=%ldk)\n", Len, Len/1024);
  for(Address = 0; Address < Len; Address++) *Buffer++ = ReadDataMemByte(Address);
  return(Len);
}    

/*--------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static int WriteDataMemByte(long Address, BYTE_T Data) {
  BYTE_T VerifyData;
  int Retry;
  Stk200SendByte(CmdWriteDataMem[0]);
  Stk200SendByte(CmdWriteDataMem[1]|((Address >> 8) & 0xFF));
  Stk200SendByte(CmdWriteDataMem[2]|( Address       & 0xFF));
  Stk200SendByte(Data);
  NanoDelay(BYTE_PRG_WAIT);
  for(Retry = 0; Retry < PRG_DATA_ROM_RETRYS; Retry++) {
    if((VerifyData = ReadDataMemByte(Address)) == Data) {
      DBG_PRINTF(DBG_IO, "WriteDataMemByte(): needed %d retries for verifing\n",\
    		Retry + 1);
      return(0);
    }
  }
  fprintf(stderr,"WriteDataMemByte(): Cannot verify Address %ld, "\
  		"wrote 0x%02X, read 0x%02X\n", Address, Data, VerifyData);
  return(-1);
}

/*------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int WriteDataMem(unsigned char *Buffer, long Len) {
  long Address;
  DBG_PRINTF(DBG_IO, "WriteDataMem(%ld=%ldk)\n", Len, Len/1024);
  for(Address = 0; Address < Len; Address++) {
    if(*Buffer != BUFFER_INIT_VAL) {
      if(WriteDataMemByte(Address, *Buffer) != 0) return(-1);
    } /* if(*Buffer != BUFFER_INIT_VAL) */
    Buffer++;
  } /* for(Address = 0; Address < Len; Address++) */
  return(Len);
}

/*------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int WriteLockAvr(LOCKBIT_T *Lock) {
  static LOCK_T LockBits[] =	{
  	{ LOCK1,	(1<<1),	},
  	{ LOCK2,	(1<<2),	},
  	/*------------ end ------------*/
  	{ NOLOCK,	0,	},	};
  int i;	
  BYTE_T LockMask = 0;
  
  if(*Lock == NOLOCK) return(-1);
  for(; *Lock != NOLOCK; Lock++) {
    for(i = 0; LockBits[i].Lb != NOLOCK; i++) {
      if(*Lock == LockBits[i].Lb) { 
        LockMask |= LockBits[i].Mask;
        break;
      } /* if(*Lock == LockBits[i].Lb) */
    } /* for(i = 0; LockBits[i].Lb != NOLOCK; i++) */
  } /* for(; *Lock != NOLOCK; Lock++) */	
  Stk200SendByte(CmdWriteLock[0]);
  Stk200SendByte(CmdWriteLock[1] & ~LockMask);
  Stk200SendByte(CmdWriteLock[2]);
  Stk200SendByte(CmdWriteLock[3]);
  return(0);
}

/*------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ReadLockAvr(LOCKBIT_T *Lock) {
  static LOCK_T LockBits[] =	{
  	{ LOCK1,	(1<<1),	},
  	{ LOCK2,	(1<<2),	},
  	/*------------ end ------------*/
  	{ NOLOCK,	0,	},	};
  int i;	
  BYTE_T LockMask;
  Stk200SendByte(CmdReadLock[0]);
  Stk200SendByte(CmdReadLock[1]);
  Stk200SendByte(CmdReadLock[3]);
  LockMask = Stk200SendByte(CmdReadLock[3]);  
  for(i = 0; LockBits[i].Lb != NOLOCK; i++) {
    if((LockMask & LockBits[i].Mask) == 0) *Lock++ = LockBits[i].Lb;
  } /* for(i = 0; LockBits[i].Lb != NOLOCK; i++) */
  *Lock = NOLOCK;
  return(0);
}

/*------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ReadLock2323_2343_4434_8535(LOCKBIT_T *Lock) {
  static LOCK_T LockBits[] =	{
  	{ LOCK1,	(1<<7),	},
  	{ LOCK2,	(1<<6),	},
  	/*------------ end ------------*/
  	{ NOLOCK,	0,	},	};
  int i;	
  BYTE_T LockMask;
  Stk200SendByte(CmdReadLock[0]);
  Stk200SendByte(CmdReadLock[1]);
  Stk200SendByte(CmdReadLock[3]);
  LockMask = Stk200SendByte(CmdReadLock[3]);  
  for(i = 0; LockBits[i].Lb != NOLOCK; i++) {
    if((LockMask & LockBits[i].Mask) == 0) *Lock++ = LockBits[i].Lb;
  } /* for(i = 0; LockBits[i].Lb != NOLOCK; i++) */
  *Lock = NOLOCK;
  return(0);
}


/*-------------------------------------------------------------------*/
static FUSE_T WrFuseBits2333_4433[] =	{
  	{ CKSEL0,	(1<<0),	},
  	{ CKSEL1,	(1<<1),	},
  	{ CKSEL2,	(1<<2),	},
  	{ BODEN,	(1<<3),	},
  	{ BODLEVEL,	(1<<4),	},
  	/*------------ end ------------*/
  	{ NOFUSE,	0,	},	};

static FUSE_T WrFuseBits2323_4434_8535[] = {
  	{ FSTRT,	(1<<0),	},
  	/*------------ end ------------*/
  	{ NOFUSE,	0,	},	};

static FUSE_T WrFuseBits2343[] =	{
  	{ RCEN,		(1<<0),	},
  	/*------------ end ------------*/
  	{ NOFUSE,	0,	},	};

/*------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static int WriteFuse(FUSEBIT_T *Fuse, FUSE_T *WrFuseBits) {
  int i;	
  BYTE_T FuseMask = 0;
  
  for(; *Fuse != NOFUSE; Fuse++) {
    for(i = 0; WrFuseBits[i].Fs != NOFUSE; i++) {
      if(*Fuse == WrFuseBits[i].Fs) { 
        FuseMask |= WrFuseBits[i].Mask;
        break;
      } /* if(*Fuse == WrFuseBits[i].Fs) */
    } /* for(i = 0; WrFuseBits[i].Fs != NOFUSE; i++) */
  } /* for(; *Fuse != NOFUSE; Fuse++) */	
  Stk200SendByte(CmdWriteFuse[0]);
  Stk200SendByte(CmdWriteFuse[1] & ~FuseMask);
  Stk200SendByte(CmdWriteFuse[2]);
  Stk200SendByte(CmdWriteFuse[3]);
  return(0);
}

/*------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int WriteFuse2333_4433(FUSEBIT_T *Fuse) {
  return(WriteFuse(Fuse, WrFuseBits2333_4433));
}


/*------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int WriteFuse2323_4434_8535(FUSEBIT_T *Fuse) {
  return(WriteFuse(Fuse, WrFuseBits2323_4434_8535));
}

/*------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int WriteFuse2343(FUSEBIT_T *Fuse) {
  return(WriteFuse(Fuse, WrFuseBits2343));
}

/*-------------------------------------------------------------------*/
static FUSE_T RdFuseBits2333_4433[] =	{
  	{ CKSEL0,	(1<<0),	},
  	{ CKSEL1,	(1<<1),	},
  	{ CKSEL2,	(1<<2),	},
  	{ BODEN,	(1<<3),	},
  	{ BODLEVEL,	(1<<4),	},
  	{ SPIEN,	(1<<5),	},
  	/*------------ end ------------*/
  	{ NOFUSE,	0,	},	};

static FUSE_T RdFuseBits2323_4434_8535[] = {
  	{ FSTRT,	(1<<0),	},
  	{ SPIEN,	(1<<5), },
  	/*------------ end ------------*/
  	{ NOFUSE,	0,	},	};

static FUSE_T RdFuseBits2343[] =	{
  	{ RCEN,		(1<<0),	},
  	{ SPIEN,	(1<<5), },
  	/*------------ end ------------*/
  	{ NOFUSE,	0,	},	};


/*------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static int ReadLockFuse(FUSEBIT_T *Fuse, FUSE_T *RdFuseBits) {
  int i;	
  BYTE_T LockMask;
  Stk200SendByte(CmdReadLock[0]);
  Stk200SendByte(CmdReadLock[1]);
  Stk200SendByte(CmdReadLock[3]);
  LockMask = Stk200SendByte(CmdReadLock[3]);  
  for(i = 0; RdFuseBits[i].Fs != NOFUSE; i++) {
    if((LockMask & RdFuseBits[i].Mask) == 0) *Fuse++ = RdFuseBits[i].Fs;
  } /* for(i = 0; RdFuseBits[i].Fs != NOFUSE; i++) */
  *Fuse = NOFUSE;
  return(0);
}

/*------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static int ReadFuse(FUSEBIT_T *Fuse, FUSE_T *RdFuseBits) {
  int i;	
  BYTE_T FuseMask;
  Stk200SendByte(CmdReadFuse[0]);
  Stk200SendByte(CmdReadFuse[1]);
  Stk200SendByte(CmdReadFuse[3]);
  FuseMask = Stk200SendByte(CmdReadFuse[3]);  
  for(i = 0; RdFuseBits[i].Fs != NOFUSE; i++) {
    if((FuseMask & RdFuseBits[i].Mask) == 0) *Fuse++ = RdFuseBits[i].Fs;
  } /* for(i = 0; RdFuseBits[i].Fs != NOFUSE; i++) */
  *Fuse = NOFUSE;
  return(0);
}

/*-------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ReadFuse2333_4433(FUSEBIT_T *Fuse) {
  return(ReadFuse(Fuse, RdFuseBits2333_4433));
}

/*-------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ReadFuse2323_4434_8535(FUSEBIT_T *Fuse) {
  return(ReadLockFuse(Fuse, RdFuseBits2323_4434_8535));
}

/*-------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int ReadFuse2343(FUSEBIT_T *Fuse) {
  return(ReadLockFuse(Fuse, RdFuseBits2343));
}
